/*
** Module   : ChangeBroker.java
** Abstract : Clearinghouse for DMO property change events
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20060719   @28114 Created initial version. Brokers notification
**                           of DMO state changes to registered listeners.
** 002 ECF 20060722   @28134 Fixed listener registration defect. Listeners
**                           registered at inner scopes were hiding those
**                           registered for the same DMO types at outer
**                           scopes.
** 003 ECF 20060802   @28348 Make addListener() tolerant of multiple calls
**                           within the same scope. Listeners are now
**                           stored in a HashSet, so duplicates are
**                           prevented.
** 004 ECF 20060918   @29665 Added forcePendingFlush() method. Forces a
**                           pending session flush for a particular DMO
**                           type.
** 005 ECF 20060919   @29696 Fixed handling of pending flush state. The
**                           pending flush state was being tracked
**                           globally, across all databases, rather than
**                           per database. This led to errors in proper
**                           detection of whether a session flush was
**                           needed, because the state was being cleared
**                           too aggressively.
** 006 ECF 20061010   @30308 Made changes to accommodate auto-commit temp
**                           tables. These were primarily side effects of
**                           changes to the Persistence class to support
**                           this feature.
** 007 ECF 20070306   @32286 Added findDirty() implementation to session
**                           interceptor. This allows us to optimize dirty
**                           checking in some cases by avoiding the
**                           default dirty check comparisons which
**                           Hibernate performs.
** 008 ECF 20070322   @32603 Added forcePendingFlush() variant. Allows
**                           forcing a pending flush when a RecordBuffer
**                           instance is not available.
** 009 ECF 20070416   @33019 Integrated user interrupt handling.
** 010 ECF 20080103   @36869 Minor change in stateChanged(). Don't force
**                           a pending flush for a record that is still
**                           transient. Also integrated generics.
** 011 ECF 20080227   @37222 Fixed SessionInterceptor.findDirty(). Was
**                           improperly short-circuiting Hiberate dirty
**                           check in non-legacy mode.
** 012 ECF 20080314   @37471 Added debug logic. Number of DMO dirty checks
**                           performed since last flush are now tracked
**                           and reported when trace level logging is
**                           enabled.
** 013 ECF 20080403   @38138 Use entity name instead of DMO class to test
**                           if flush is required.
** 014 ECF 20080822   @39554 Modified stateChanged() to accept an event
**                           type parameter. Uses inner enum Type from
**                           RecordChangeEvent.
** 015 CA  20080815   @39452 Support API change in DBUtils.
** 016 SVL 20081024   @40217 Added addListener(.., int depth) function.
** 017 ECF 20090212   @41294 Implemented eviction of unused DMOs after a
**                           session flush. Temp table DMOs may be left
**                           behind after a bulk delete, causing more
**                           flushing effort than is necessary. We now
**                           evict these objects after a flush, if they
**                           are not in use.
** 018 ECF 20090610   @42971 Minor change to SessionInterceptor.postFlush().
**                           Log database info in trace mode.
** 019 ECF 20090716   @43214 Minor change to minimize memory used by listener
**                           collections.
** 020 CA  20131013          Added no-op scopeDeleted() method, required by the changes in 
**                           Scopeable interface.
** 021 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * A clearing house which notifies interested parties of changes made to
 * active records within the current client context.  These notifications
 * are used for various purposes:
 * <ul>
 *   <li>to enable {@link RecordBuffer} to defer creating a placeholder
 *       snapshot as long as possible;
 *   <li>to permit {@link Persistence} query methods to disable auto-flushing
 *       of the active Hibernate session when possible;
 *   <li>to allow {@link AdaptiveQuery} to invalidate its current result set
 *       when necessary.
 * </ul>
 * <p>
 * {@link com.goldencode.p2j.persist.event.RecordChangeListener
 * RecordChangeListener} objects register to receive notifications for one or
 * more DMO types, using the {@link #addListener} method.  A listener will
 * only receive notifications for changes to DMOs of the types for which they
 * registered.  Beyond this, no other filtering is performed.  Listeners are
 * maintained in a scoped dictionary, so there is never a need to remove a
 * listener explicitly;  when the scope in which the listener was registered
 * is closed, it is released automatically.
 * <p>
 * Change notifications are driven by the {@link #stateChanged} method.  It
 * is anticipated that this only will be invoked by <code>RecordBuffer</code>
 * objects, as user code calls setter methods on DMO proxies.  A state change
 * causes a {@link com.goldencode.p2j.persist.event.RecordChangeEvent
 * RecordChangeEvent} to be broadcast to all listeners registered against the
 * type of DMO which was modified.  "Type" in this sense corresponds to the
 * DMO interface.
 * <p>
 * This class also enlists a Hibernate <code>Interceptor</code> to listen for
 * Hibernate session events.  The context local <code>Persistence</code>
 * instance installs this interceptor into each Hibernate <code>Session</code>
 * it opens.  Whenever a DMO state change occurs, the implementation class of
 * the modified DMO is stored in a set of pending flush types.  When the
 * interceptor is notified of a session flush event, this set is cleared.
 * As an optimization, the <code>Persistence</code> class consults with the
 * <code>ChangeBroker</code> to determine whether a session flush is required
 * before a query is executed.  If not, it disables auto-flushing for the
 * query, which preempts considerable, unnecessary work.
 * <p>
 * The interceptor has another responsibility associated with temp tables.
 * Temporary DMOs can be removed from the database via a bulk delete, which
 * bypasses the <code>BufferManager</code>'s DMO reference counting mechanism.
 * As a result, a bulk delete can leave records associated with the open
 * Hibernate session, which can result in much more flushing effort than
 * necessary, in certain situations.  During a flush event, the interceptor
 * detects DMOs which are not in use and schedules these for eviction from the
 * session.  The eviction takes place immediately after the flush event is
 * finished.
 * <p>
 * Because it relies upon the <code>RecordBuffer</code> as the source for all
 * state change events, <code>ChangeBroker</code> is not useful in non-legacy
 * client contexts (i.e., where external code accesses the services of the
 * <code>Persistence</code> class directly).  For the most part, this class
 * is transparent for these contexts.  One exception is that the auto-flush
 * optimization technique described above would be harmful if allowed to
 * proceed naturally, because flushing would always be disabled since state
 * changes would never be reported, thus it would always seem flushing is not
 * required.  To prevent this problem, this class detects when it is not in
 * legacy mode and disables this optimization accordingly in such cases.
 */
final class ChangeBroker
implements Scopeable
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(ChangeBroker.class);
   
   /** Context local instance of this class */
   private static final ContextLocal<ChangeBroker> context =
   new ContextLocal<ChangeBroker>()
   {
      @Override
      public WeightFactor getWeight() { return WeightFactor.LEVEL_5; };
      protected ChangeBroker initialValue() { return (new ChangeBroker()); }
   };
   
   /** All parties which have registered interest in DMO property changes */
   private final
   ScopedDictionary<Class<?>, Set<RecordChangeListener>> listeners =
      new ScopedDictionary<Class<?>, Set<RecordChangeListener>>();
   
   /** Objects which are notified of activity in Hibernate session */
   private final Map<Persistence, SessionInterceptor> interceptors =
      new IdentityHashMap<Persistence, SessionInterceptor>();
   
   /** Is this client context using legacy services? */
   private boolean legacyMode = false;
   
   /**
    * This class can only be constructed privately.
    */
   private ChangeBroker()
   {
   }
   
   /**
    * Retrieve the context-local instance of this class, instantiating it
    * first if necessary.
    *
    * @return  Context-local instance of this class.
    */
   static ChangeBroker get()
   {
      return context.get();
   }
   
   /**
    * Register with the {@link com.goldencode.p2j.util.TransactionManager
    * TransactionManager} a factory object which creates instances of this
    * class, so that they can be registered to receive notifications of
    * runtime scope start and finish events.
    * <p>
    * This method should be invoked once during the server bootstrap phase.
    */
   static void initialize()
   {
      ScopeableFactory factory = new ScopeableFactory()
      {
         public Scopeable createScopeable()
         {
            return ChangeBroker.get();
         }
      };
      TransactionManager.registerScopeableFactory(factory);
   }
   
   /**
    * Add a scope to the listeners dictionary whenever a new runtime scope
    * opens.
    */
   public void scopeStart()
   {
      listeners.addScope(null);
   }
   
   /**
    * Pop the current scope from the listeners dictionary whenever a runtime
    * scope is closed.
    */
   public void scopeFinished()
   {
      listeners.deleteScope();
   }
   
   /**
    * Notification that the instantiating procedure where the scope was opened was deleted.
    * <p>
    * This implementation is a no-op.
    */
   @Override
   public void scopeDeleted()
   {
      // no-op
   }
   
   /**
    * Register a listener to receive DMO property change notifications.  The
    * listener will receive notifications of any changes to DMOs of the types
    * for which it is registered, as they occur in this scope or in scopes
    * nested within this scope.  The listener is deregistered automatically
    * when the scope ends.
    * <p>
    * This method is tolerant of being called multiple times for the same
    * listener in the same scope;  the listener will only be registered once,
    * and therefore will receive only one notification per distinct event.
    *
    * @param   listener
    *          Listener object to be added.
    */
   void addListener(RecordChangeListener listener)
   {
      addListener(listener, listeners.size());
   }
   
   /**
    * Register a listener to receive DMO property change notifications. The
    * registration is performed in the current scope and the outer scopes,
    * which number is specified by the <code>depth</code> parameter. The
    * listener will receive notifications of any changes to DMOs of the types
    * for which it is registered, as they occur in the specified scopes or in
    * the nested ones.  The listener is deregistered automatically
    * when the scopes end.
    * <p>
    * When we call this function with a <code>depth</code> less than the
    * number of innermost scope, we actually want to register it in some outer
    * scope, but since registrations at inner scopes will hide those at outer
    * scopes during a lookup, we should register it in the several scopes
    * if needed.
    * <p>
    * This method is tolerant of being called multiple times for the same
    * listener in the same scope;  the listener will only be registered once,
    * and therefore will receive only one notification per distinct event.
    *
    * @param   listener
    *          Listener object to be added.
    *
    * @param   depth
    *          1-based number of the most outer scope in which we want to
    *          register the listener (starting from the very outermost scope).
    */
   void addListener(RecordChangeListener listener, int depth)
   {
      legacyMode = true;
      
      Iterator<RecordBuffer> iter = listener.recordBuffers();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next();
         Class<?> dmoIface = buffer.getDMOInterface();
         Set<RecordChangeListener> interested =
            listeners.getValueAtScope(dmoIface, 0);
         if (interested == null)
         {
            interested = new LinkedHashSet<RecordChangeListener>(4);
            
            // If there are any existing listeners at outer scopes, we need
            // to add them all to this list, because this list will hide
            // them at lookup time.  They are added first, so they will be
            // notified first.
            Set<RecordChangeListener> existingListeners =
               listeners.lookup(dmoIface);
            if (existingListeners != null)
            {
               interested.addAll(existingListeners);
            }
            listeners.addEntry(false, dmoIface, interested);
         }
         interested.add(listener);
         
         int depthDiff = listeners.size() - depth;
         for (int i = 1; i <= depthDiff; i++)
         {
            interested = listeners.getValueAtScope(dmoIface, i);
            if (interested != null)
            {
               interested.add(listener);
            }
         }
      }
   }
   
   /**
    * Force a Hibernate session flush to occur before the next query against
    * the given buffer is executed.
    *
    * @param   buffer
    *          Buffer whose DMO type must be registered for a pending flush.
    */
   void forcePendingFlush(RecordBuffer buffer)
   {
      forcePendingFlush(buffer.getPersistence(),
                        buffer.getDMOImplementationClass());
   }
   
   /**
    * Force a Hibernate session flush to occur before the next query against
    * the given DMO class is executed.
    *
    * @param   persistence
    *          Persistence service object.
    * @param   dmoClass
    *          DMO implementation class for which flushing must occur.
    */
   void forcePendingFlush(Persistence persistence, Class<?> dmoClass)
   {
      SessionInterceptor interceptor = getInterceptor(persistence);
      interceptor.setFlushPending(dmoClass.getName());
   }
   
   /**
    * Broadcast a notification that a DMO's state has changed.  All listeners
    * registered to receive events about DMOs of the type managed by
    * <code>buffer</code> will receive the event.
    *
    * @param   type
    *          Event type:  insert, update, or delete.
    * @param   buffer
    *          Record buffer which originated the event.
    * @param   dmo
    *          <b>Post-modification</b> state of the affected DMO (in the
    *          case where the modification is the deletion of the DMO from
    *          the database, this will be the object instance which was
    *          deleted).
    * @param   snapshot
    *          A deep copy of the DMO which reflects the state of the DMO as
    *          it was first set into the record buffer.  For a newly created
    *          record, this will contain the primary key assigned to the
    *          record.  All other properties will be set to their default
    *          values.
    * @param   properties
    *          A map whose keys are the names of those properties which have
    *          been modified and whose values are lists of the indexes at
    *          which those values are stored (indexed properties only).  For
    *          a simple property, the matching value will be
    *          <code>null</code>.
    *
    * @throws  PersistenceException
    *          if any error occurs within a listener while processing a
    *          change event.  Exceptions will be deferred until all listeners
    *          have had a chance to process the event.  If more than one
    *          listener throws an exception, only the first will be rethrown
    *          by this method;  however, all will be logged.
    */
   void stateChanged(RecordChangeEvent.Type type,
                     RecordBuffer buffer,
                     Persistable dmo,
                     Persistable snapshot,
                     Map<String, List<Integer>> properties)
   throws PersistenceException
   {
      // If the buffer is not transient, then the change represents either an
      // insert or an update/delete of an existing record.  In these cases, we
      // want to ensure the database is synchronized at the next opportunity.
      if (!buffer.isTransient())
      {
         forcePendingFlush(buffer);
      }
      
      Class<?> dmoIface = buffer.getDMOInterface();
      Collection<RecordChangeListener> interested =
         listeners.lookup(dmoIface);
      if (interested == null)
      {
         return;
      }
      
      RecordChangeEvent event = new RecordChangeEvent(type,
                                                      buffer,
                                                      dmo,
                                                      snapshot,
                                                      properties);
      
      Exception error = null;
      
      // Notify all registered listeners.
      Iterator<RecordChangeListener> iter = interested.iterator();
      while (iter.hasNext())
      {
         RecordChangeListener listener = iter.next();
         try
         {
            listener.stateChanged(event);
         }
         catch (Exception exc)
         {
            DBUtils.handleException(buffer.getDatabase(), exc);
            
            // Can only throw the first exception caught, but we want to
            // give all listeners a chance to process the event.
            if (error == null)
            {
               error = exc;
            }
            
            // Log it regardless.
            LOG.error(exc);
         }
      }
      
      if (error != null)
      {
         throw new PersistenceException(error);
      }
   }
   
   /**
    * Report whether a Hibernate session level flush would be required before
    * processing a query which expects results of type <code>dmoClass</code>.
    * A flush is required if any change has been made to a DMO of this type
    * in the current context, since the last time a session flush occurred.
    * If <code>dmoClass</code> is <code>null</code>, this method determines
    * that a flush is required if a DMO of <i>any</i> type has been modified
    * since the last time a session flush occurred.
    * <p>
    * This method will always recommend a flush when executing in a
    * non-legacy client context.
    *
    * @param   persistence
    *          Persistence object which is the key under which the
    *          interceptor object is stored.
    * @param   entityName
    *          DMO implementation class for which to check for pending flush,
    *          or <code>null</code> to check for any class.
    *
    * @return  <code>true</code> if a flush is required, else
    *          <code>false</code>.
    */
   boolean requiresFlush(Persistence persistence, String entityName)
   {
      if (!legacyMode)
      {
         return true;
      }
      
      SessionInterceptor interceptor = getInterceptor(persistence);
      
      return interceptor.isFlushPending(entityName);
   }
   
   /**
    * Retrieve the <code>Interceptor</code> instance to be installed in the
    * next Hibernate <code>Session</code> to be opened against the specified
    * database.
    *
    * @param   persistence
    *          Persistence object which is the key under which the
    *          interceptor object is stored.
    *
    * @return  This broker's <code>SessionInterceptor</code> instance.
    */
   SessionInterceptor getInterceptor(Persistence persistence)
   {
      SessionInterceptor interceptor = interceptors.get(persistence);
      if (interceptor == null)
      {
         interceptor = new SessionInterceptor(persistence);
         interceptors.put(persistence, interceptor);
      }
      
      return interceptor;
   }
   
   /**
    * An implementation of a Hibernate <code>Interceptor</code> which listens
    * for session flush events and clears the enclosing instance's pending
    * flush status whenever a flush occurs.
    */
   private class SessionInterceptor
   extends EmptyInterceptor
   {
      /** Persistence service object associated with this interceptor */
      private final Persistence persistence;
      
      /** Indicates an entity is not dirty */
      private final int[] NOT_DIRTY = new int[] {};
      
      /** Set of DMO entities which require a Hibernate session flush */
      private final Set<String> pendingFlushTypes = new HashSet<String>();
      
      /** Flag indicating trace-level debug logging */
      private final boolean trace = LOG.isTraceEnabled();
      
      /** Set of (temp table) DMOs in active use for the associated session */
      private final Set<Persistable> recordsInUse;
      
      /** Number of dirty checks performed by Hibernate since last flush */
      private int dirtyChecks = 0;
      
      /** Set of records detecting during flush as not being in active use */
      private Set<Persistable> pendingEvictions = new HashSet<Persistable>();
      
      /** Flag to track presence of any DMO entities which need flushing */
      private boolean empty = true;
      
      /**
       * Constructor.
       * 
       * @param   persistence
       *          Persistence service object associated with this interceptor.
       */
      private SessionInterceptor(Persistence persistence)
      {
         this.persistence = persistence;
         
         if (DatabaseManager.TEMP_TABLE_DB.equals(persistence.getDatabase()))
         {
            // Store a view of the temp table DMOs which the buffer manager is
            // tracking as currently in use.  This allows us to account for
            // DMOs which may be left in the Hibernate session after a bulk
            // delete, so that we can evict them after a flush event.
            BufferManager bufMan = BufferManager.get();
            recordsInUse = bufMan.getInUseDMOs(persistence);
         }
         else
         {
            // Don't track in-use DMOs for permanent tables.  In the current
            // implementation, there is no bulk delete feature for permanent
            // tables, so we can be pretty sure that the only records in the
            // Hibernate session are those that we want there.
            recordsInUse = null;
         }
      }
      
      /**
       * Indicates whether a data model object entity either:
       * <ul>
       *   <li>might be dirty, possibly requiring a flush to the database;  or
       *   <li>definitely is not dirty.
       * </ul>
       * <p>
       * We do not definitively indicate that a record <i>is</i> dirty, since
       * Hibernate does not honor this indication when it comes to composite
       * elements within first class entities.  Therefore, when we can detect
       * that an entity might be dirty (i.e., its DMO class has been added to
       * the set of pending flush types), we return <code>null</code> to
       * indicate to Hibernate that it should use its default dirty checking
       * strategy to make this determination for sure.
       * <p>
       * We can, however, definitively detect when an entity cannot be dirty
       * (i.e., its DMO class is not present in the set of pending flush
       * types).  In this case, we short-circuit Hibernate's default checking
       * as an optimization.
       *
       * @param   entity
       *          The DMO to be tested for modifications.
       * @param   id
       *          The DMO's primary key ID.
       * @param   currentState
       *          Array of all of the DMO's properties in their current state.
       * @param   previousState
       *          Array of all of the DMO's properties in their previous
       *          state, presumably before they were updated.
       * @param   propertyNames
       *          Property names corresponding with the previous two arrays
       *          parameters.
       * @param   types
       *          Hibernate types corresponding with the previous three
       *          array parameters.
       *
       * @return  An empty array of <code>int</code>s to indicate that the
       *          entity is not dirty, or <code>null</code> to indicate the
       *          entity may be dirty.  Always return <code>null</code> if
       *          not in legacy mode, because we won't have enough information
       *          to second-guess Hibernate in that case.
       */
      public int[] findDirty(Object entity,
                             Serializable id,
                             Object[] currentState,
                             Object[] previousState,
                             String[] propertyNames,
                             Type[] types)
      {
         if (trace)
         {
            dirtyChecks++;
         }
         
         // This optimization is only valid in legacy mode, because we don't
         // track changes in non-legacy mode.
         if (!legacyMode)
         {
            return null;
         }
         
         // If the record is not in use, prepare to evict it after the flush.
         if (recordsInUse != null)
         {
            Persistable dmo = (Persistable) entity;
            if (!recordsInUse.contains(dmo))
            {
               pendingEvictions.add(dmo);
            }
         }
         
         if (empty)
         {
            // Short circuit dirty check.
            return NOT_DIRTY;
         }
         
         Class<?> dmoClass = entity.getClass();
         dmoClass = DatabaseManager.normalizeDMOClass(dmoClass);
         if (!isFlushPending(dmoClass.getName()))
         {
            // Short circuit dirty check.
            return NOT_DIRTY;
         }
         
         // Let Hibernate figure it out.
         return null;
      }
      
      /**
       * Invoked by Hibernate after the current session has been flushed so
       * as to execute the SQL statements required to synchronize in-memory
       * session state with the database.  Triggers a full reset of the
       * pending flush state for this interceptor's database, such that the
       * broker will not recommend a session flush before the next DMO change
       * occurs.
       *
       * @param   entities
       *          An iterator on the DMO instances whose updates were flushed
       *          to the database.
       */
      public void postFlush(Iterator entities)
      {
         if (trace)
         {
            StringBuilder buf = new StringBuilder();
            buf.append("[");
            buf.append(Utils.describeContext());
            buf.append("] ");
            buf.append(persistence.getDatabase());
            buf.append(" dirty check count:  ");
            buf.append(dirtyChecks);
            LOG.trace(buf.toString());
            dirtyChecks = 0;
         }
         
         cleanup();
      }
      
      /**
       * Called after a transaction is committed or rolled back.  If the
       * transaction was rolled back, triggers a full reset of the pending
       * flush state for this interceptor's database, such that the broker
       * will not recommend a session flush before the next DMO change
       * occurs.
       */
      public void afterTransactionCompletion(Transaction tx)
      {
         if (tx.wasRolledBack())
         {
            cleanup();
         }
      }
      
      /**
       * Indicate whether a Hibernate session flush is pending for the given
       * DMO implementation class.
       *
       * @param   entityName
       *          DMO implementation class as stored by the DMO's record
       *          buffer.  Do not use <code>getClass()</code> on a DMO to
       *          retrieve this argument, as Hibernate may have proxied the
       *          DMO and an unexpected class will be returned.
       *
       * @return  <code>true</code> if a flush is pending for the DMO type
       *          indicated by <code>dmoClass</code>.  If the argument is
       *          <code>null</code>, this method conservatively returns
       *          <code>true</code> if <i>any</i> DMO class has a session
       *          flush pending.
       */
      boolean isFlushPending(String entityName)
      {
         if (entityName != null)
         {
            return (!empty && pendingFlushTypes.contains(entityName));
         }
         
         return !empty;
      }
      
      /**
       * Set a Hibernate session flush pending for the given DMO
       * implementation class.
       *
       * @param   entityName
       *          DMO implementation class as stored by the DMO's record
       *          buffer.  Do not use <code>getClass()</code> on a DMO to
       *          retrieve this argument, as Hibernate may have proxied the
       *          DMO and an unexpected class will be returned.
       */
      void setFlushPending(String entityName)
      {
         pendingFlushTypes.add(entityName);
         empty = false;
      }
      
      /**
       * Clean up the interceptor's state after a flush or transaction.  This
       * involves clearing the pending flush types collections and evicting
       * any records we detected during the flush as not being in use.
       */
      private void cleanup()
      {
         pendingFlushTypes.clear();
         empty = true;
         if (!pendingEvictions.isEmpty())
         {
            persistence.evict(pendingEvictions);
            pendingEvictions.clear();
         }
      }
   }
}
