/*
** Module   : TempTableConnectionProvider.java
** Abstract : specialized connection provider used for temp tables
**
** Copyright (c) 2006-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 ECF 20060619 @27502   Created initial version. A JDBC connection provider implementation 
**                           used by Hibernate to hold on to a particular JDBC connection for as 
**                           long as one or more temp tables are using it.
** 002 SVL 20120331          Upgraded to Hibernate 4.
** 003 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
*/

package com.goldencode.p2j.persist;

import java.lang.InstantiationException;
import java.sql.*;
import java.util.*;

import org.apache.commons.logging.*;
import org.hibernate.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;
import org.hibernate.service.*;
import org.hibernate.service.jdbc.connections.spi.*;
import org.hibernate.service.spi.*;

/**
 * A specialized connection provider which ensures that a database connection
 * used to create temp tables is preserved until the last temp table in use
 * is dropped.  A delegate connection provider is used to do the actual work
 * of obtaining and releasing connections, presumably from a connection pool.
 * This implementation will not release a connection back to the underlying
 * delegate until there are no more temp tables using the connection.  This
 * prevents the inadvertent loss of temp table data which would otherwise
 * occur if a Hibernate session is closed (as from a commit or rollback),
 * before all temp table processing begun on that connection was finished.
 * <p>
 * Another approach to solving this problem would have been to make the
 * persistence framework responsible for obtaining database connections for
 * temp tables and to supply these manually when opening new sessions via the
 * Hibernate <code>SessionFactory.openSession(Connection connection)</code>
 * method.  However, using that approach we would have lost the ability to
 * use Hibernate's second level cache for temp table records, AND we would
 * have had to code specialized logic in {@link Persistence} to differentially
 * handle temp tables and permanent tables.  Thus, an implementation of
 * Hibernate's <code>ConnectionProvider</code> interface was the preferred
 * approach.
 * <p>
 * This implementation wrappers a delegate connection provider whose
 * properties are specified in the directory.  To enable this implementation,
 * the fully qualified name of this class (i.e.,
 * <code>com.goldencode.p2j.persist.TempTableConnectionProvider</code>) is
 * specified for the <code>hibernate.connection.provider_class</code> property
 * in the directory.  The backing provider may be specified as well, using one
 * of the following mechanisms (see also the {@link #configure} method):
 * <ul>
 * <li>Specify an explicit, fully qualified provider class name using the
 *     property <code>hibernate.connection.delegate_provider_class</code>
 *     (note this is a P2J extension, not a standard Hibernate property name).
 * <li>Do not specify an explicit delegate provider class name, but instead
 *     let Hibernate's internal logic determine the provider to initialize,
 *     based upon the remaining Hibernate configuration properties provided
 *     in the directory (see Hibernate's configuration documentation for
 *     details on how it makes this determination).
 * </ul>
 * <p>
 * This class effectively operates as a singleton, even though it is possible
 * to create multiple instances, since Hibernate requires a public, default
 * constructor.  However, the implementation consists of only static
 * variables.  One of these is a <code>ContextLocal</code> variable
 * (<code>context</code>), which maintains connection state specific to each
 * client context.  This mechanism is used to ensure that even if multiple
 * instances of this class are created, requests to get and close JDBC
 * connections are managed at a client context level.
 */
public final class TempTableConnectionProvider
implements ConnectionProvider,
           Configurable
{
   /** Logger */
   private static final Log LOG =
      Utils.getLog(TempTableConnectionProvider.class);
   
   /** Property name in configuration for connection provider class name */
   private static final String PROVIDER_CLASS_KEY =
      "hibernate.connection.provider_class";
   
   /** Property name in configuration for delegate provider class name */
   private static final String DELEGATE_CLASS_KEY =
      "hibernate.connection.delegate_provider_class";
   
   /** Object which manages context local connection data */
   private static final ContextLocal context = new ContextLocal()
   {
      /**
       * Get the weight of this context-local var.
       * 
       * @return   Always {@link WeightFactor#LEVEL_2}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_2;
      };

      /**
       * Instantiate a new <code>Context</code> instance for each context.
       *
       * @return  Context object.
       */
      protected Object initialValue()
      {
         return new Context();
      }
      
      /**
       * Clean up the context before it is removed.
       *
       * @param   value
       *          Context object stored in this variable.
       */
      protected void cleanup(Object value)
      {
         if (value != null)
         {
            ((Context) value).contextCleanup();
         }
      }
   };
   
   /** Connection provider to which actual work is delegated */
   private static ConnectionProvider delegate = null;
   
   /**
    * Default ctor used by Hibernate's <code>ConnectionProviderFactory</code>
    * to instantiate this connection provider.
    */
   public TempTableConnectionProvider()
   {
   }
   
   /**
    * Configure this connection provider using the specified properties.
    * It is here where we construct and configure the delegate connection
    * provider.  The properties are for the most part passed through to the
    * connection provider factory to produce the delegate object.  However,
    * one property in particular is modified or removed.  The
    * <code>hibernate.connection.provider_class</code> property either:
    * <ol>
    * <li>has its value replaced with the value of the
    *     <code>hibernate.connection.delegate_provider_class</code> (if
    *     provided);  or
    * <li>is removed altogether if the
    *     <code>hibernate.connection.delegate_provider_class</code> property
    *     is not specified in the configuration.
    * </ol>
    * <p>
    * In the latter case, Hibernate's internal logic determines which
    * connection provider to create, based upon the remaining properties.
    *
    * @param   props
    *          Hibernate configuration properties.
    */
   public void configure(Map props)
   {
      if (delegate != null)
      {
         return;
      }
      
      // Copy all properties to pass to the connection provider factory when
      // requesting a delegate provider.
      Properties propCopy = new Properties();
      propCopy.putAll(props);
      
      // If a delegate connection provider class is specified, use it to
      // initialize the delegate.  Otherwise, remove the provider_class
      // property from the properties and try to create a new connection
      // provider instance as the delegate.
      String delegateClass = (String) props.get(DELEGATE_CLASS_KEY);
      if (delegateClass != null)
      {
         // Provide an explicit provider class for the delegate.
         propCopy.setProperty(PROVIDER_CLASS_KEY, delegateClass);
      }
      else
      {
         // Let Hibernate figure out which provider to create, based upon the
         // remaining properties.
         propCopy.remove(PROVIDER_CLASS_KEY);
      }
      
      // Instantiate and configure the proper backing provider.
      String cls = propCopy.getProperty(PROVIDER_CLASS_KEY);
      try
      {
         delegate = (ConnectionProvider) Class.forName(cls).newInstance();
         ((Configurable) delegate).configure(propCopy);
      }
      catch (InstantiationException e)
      {
         throw new HibernateException("Cannot instantiate connection provider " + cls);
      }
      catch (IllegalAccessException e)
      {
         throw new HibernateException("Cannot access connection provider " + cls);
      }
      catch (ClassNotFoundException e)
      {
         throw new HibernateException("Cannot find connection provider " + cls);
      }

      if (LOG.isInfoEnabled())
      {
         LOG.info("Temp table connection provider configured;  backed by " +
                  delegate.getClass().getName());
      }
   }
   
   /**
    * Obtain a JDBC connection on demand.  For a particular client context,
    * this method will fetch a new connection from the underlying delegate
    * provider only if no connection has previously been fetched, or if a
    * previously fetched connection was closed after all temp tables using it
    * were dropped.
    *
    * @return  Active temp table JDBC connection for the current context.
    *
    * @throws  SQLException
    *          if there is an error establishing a connection with the
    *          backing datastore.
    */
   public Connection getConnection()
   throws SQLException
   {
      return ((Context) context.get()).getConnection();
   }
   
   /**
    * Return a JDBC connection which is no longer needed in the current
    * context, releasing it to the underlying delegate provider if it is no
    * longer needed by any active temp tables.  If any temp table still
    * references it, it is not released to the delegate.
    *
    * @param   conn
    *          The connection to be released.
    *
    * @throws  SQLException
    *          if there is an error closing the connection with the backing
    *          datastore.
    */
   public void closeConnection(Connection conn)
   throws SQLException
   {
      ((Context) context.get()).closeConnection(conn);
   }
   
   /**
    * Returns <code>false</code> to indicate JDBC connections will not be
    * released aggressively, regardless of how the delegate would handle such
    * requests.
    *
    * @return  <code>false</code>.
    */
   public boolean supportsAggressiveRelease()
   {
      return false;
   }

   /**
    * Can this wrapped service be unwrapped as the indicated type?
    *
    * @param  unwrapType
    *         The type to check.
    *
    * @return <code>true</code>/<code>false</code>
    */
   public boolean isUnwrappableAs(Class unwrapType)
   {
      return ConnectionProvider.class.equals(unwrapType) ||
				 TempTableConnectionProvider.class.isAssignableFrom(unwrapType);
   }

   /**
    * Unproxy the service proxy.
    *
    * @param  unwrapType
    *         The java type as which to unwrap this instance.
    *
    * @return The unwrapped reference.
    *
    * @throws UnknownUnwrapTypeException if the service cannot be unwrapped as the indicated type.
    */
   public <T> T unwrap(Class<T> unwrapType)
   {
      if (isUnwrappableAs(unwrapType))
      {
			return (T) this;
		}
		else
      {
			throw new UnknownUnwrapTypeException(unwrapType);
		}
   }

   /**
    * An instance of this class is created for each client context, in order
    * to manage connection data specific to that context.  This class
    * determines when the delegate will be asked to obtain and release the
    * current JDBC connection, using a reference count maintained by the
    * {@link TemporaryBuffer} class.
    */
   private static class Context
   {
      /** Active JDBC connection for temp tables in this client context */
      private Connection connection = null;
      
      /**
       * Fetch a JDBC connection, which will be either the cached connection
       * (if previously opened and not subsequently closed), or which will be
       * newly obtained from the delegate connection provider (if it is not
       * currently cached).
       *
       * @return  JDBC connection.
       *
       * @throws  SQLException
       *          if an error occurs obtaining the connection from the
       *          backing datastore.
       */
      private Connection getConnection()
      throws SQLException
      {
         if (connection == null)
         {
            connection = delegate.getConnection();
            
            if (LOG.isDebugEnabled())
            {
               LOG.debug(status("Got backing connection"));
            }
         }
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug(status("getConnection"));
         }
         
         return connection;
      }
      
      /**
       * Optionally release a JDBC connection to the delegate connection
       * provider.  The connection will be released to the delegate only if
       * no temp tables currently are using it.
       *
       * @param   conn
       *          JDBC connection to be returned to the provider.
       *
       * @throws  IllegalStateException
       *          if the connection returned (<code>conn</code>) is not the
       *          same connection currently cached by this object.
       * @throws  SQLException
       *          if an error occurs closing the connection.
       */
      private void closeConnection(Connection conn)
      throws SQLException
      {
         if (conn != connection && connection != null)
         {
            throw new IllegalStateException(
               "Attempted to close unknown temp table database connection");
         }
         
         if (LOG.isDebugEnabled())
         {
            LOG.debug(status("closeConnection"));
         }
         
         if (connection != null && TemporaryBuffer.getTableCount() == 0)
         {
            delegate.closeConnection(connection);
            connection = null;
         
            if (LOG.isDebugEnabled())
            {
               LOG.debug(status("Released backing connection"));
            }
         }
      }
      
      /**
       * Perform clean up when the current client context is being closed.
       * If a connection currently is cached, is returned to the delegate
       * provider, so that it may be recycled or released.
       */
      private void contextCleanup()
      {
         if (connection == null)
         {
            return;
         }
         
         try
         {
            delegate.closeConnection(connection);
         }
         catch (SQLException exc)
         {
            LOG.error("Error closing JDBC connection", exc);
         }
      }
      
      /**
       * Generate a debug string which reports the current status of this
       * object along with a custom message.  Status includes the number of
       * temp tables currently accessing the cached connection, as well as
       * an identifier of the cached connection itself (if any).
       *
       * @param   message
       *          Message to be prepended to the status string.
       *
       * @return  Text which reports the given message plus the current
       *          status of this object.
       */
      private String status(String message)
      {
         StringBuffer buf = new StringBuffer(Utils.describeContext());
         buf.append(":  ");
         buf.append(message);
         buf.append(";  tableCount: ");
         buf.append(TemporaryBuffer.getTableCount());
         buf.append(";  connection: ");
         buf.append(connection);
         
         return buf.toString();
      }
   }
}
