/*
** Module   : FieldAssigner.java
** Abstract : Assigns [input-]output parameter values to database fields upon
**            exit from procedure/function
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 ECF 20090816   @43673 Created initial version. Assigns [input-]output parameter values to 
**                           database fields upon exit from procedure/function.
** 002 CA  20131013          Added no-op scopeDeleted() method, required by the changes in 
**                           Scopeable interface.
** 003 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
*/

package com.goldencode.p2j.persist.type;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * Delegate worker for the various <code>*Field</code> data type wrapper
 * implementations, which assigns an input-output or output parameter's value
 * back to the database field with which it is associated, upon exit from a
 * procedure or function.
 * <p>
 * An individual instance of this class is created by each <code>*Field</code>
 * instance, and that variable is stored within this object.  At construction,
 * it initializes the parameter (if necessary) to the appropriate current
 * value from the database.  Upon exit from the procedure or function with
 * which it is associated, it attempts to assign the current value of its
 * contained variable back into the associated DMO.  This may trigger normal
 * validation processing, which may produce error conditions.
 */
public final class FieldAssigner
{
   /** Object which manages all instances of this class by context */
   private static final ContextLocal<Context> context =
   new ContextLocal<Context>()
   {
      @Override
      public WeightFactor getWeight() { return WeightFactor.LEVEL_5; };
      protected Context initialValue() { return new Context(); }
   };
   
   /** Data wrapper which is used as a procedure or function parameter */
   private final BaseDataType variable;
   
   /** Field reference which manages the association with the DMO property */
   private final FieldReference field;
   
   /**
    * Constructor.
    * 
    * @param   variable
    *          Parameter value to and from which we will assign.
    * @param   dmo
    *          Proxy for record whose field is being accessed.
    * @param   property
    *          Name of dmo field.
    * @param   index
    *          Index to access, if property represents an extent field.
    * @param   initialize
    *          Should this variable be initialized with the DMO's current
    *          value for the target field?
    */
   FieldAssigner(BaseDataType variable,
                 DataModelObject dmo,
                 String property,
                 NumberType index,
                 boolean initialize)
   {
      this.variable = variable;
      this.field = new FieldReference(dmo, property, index);
      
      if (initialize)
      {
         variable.assign(field.get());
      }
      
      Context local = context.get();
      Scope scope = local.getCurrentScope();
      scope.addAssigner(this);
   }
   
   /**
    * Retrieve a <code>Scopeable</code> for the current context.  This allows
    * an external object to notify this class' current context of scope open
    * and close events.
    * 
    * @return  Context local <code>Scopeable</code>.
    */
   public static Scopeable getScopeable()
   {
      return context.get();
   }
   
   /**
    * Assign the current value of the parameter variable back to the
    * associated DMO property.  This is expected to occur upon exit from the
    * function or procedure with which this object is associated.
    */
   private void assign()
   {
      field.set(variable);
   }
   
   /**
    * Object which manages the context local state of field assigners.  It
    * manages a deque of scopes, each of which contains zero or more
    * <code>FieldAssigner</code> objects.  As the currently executing
    * program's block scopes open and close, these scopes are pushed and
    * popped from the deque.
    */
   private static class Context
   implements Scopeable
   {
      /** Deque containing field assigners by scope */
      private final ArrayDeque<Scope> scopes = new ArrayDeque<Scope>();
      
      /**
       * Default constructor.
       */
      private Context()
      {
         scopes.push(new Scope());
      }
      
      /**
       * Called as each block scope opens, at which time we push a scope onto
       * the deque.
       * 
       * @see com.goldencode.p2j.util.Scopeable#scopeStart()
       */
      public void scopeStart()
      {
         scopes.push(new Scope());
      }
      
      /**
       * Called as each block scope closes, at which time we pop a scope from
       * the deque.
       * 
       * @see com.goldencode.p2j.util.Scopeable#scopeFinished()
       */
      public void scopeFinished()
      {
         scopes.pop();
      }
      
      /** No-op. */
      @Override
      public void scopeDeleted()
      {
         // no-op
      }
      
      /**
       * Get the object which manages field assigners for the current scope.
       * 
       * @return  Current scope.
       */
      private Scope getCurrentScope()
      {
         return scopes.peek();
      }
   }
   
   /**
    * One instance of this class is created for each block scope which the
    * currently executing program enters.  It contains zero or more field
    * assigners specific to that scope.
    * <p>
    * After exiting a function or procedure, the {@link #processAssignments()}
    * method is invoked, which causes each field assigner to assign the
    * current value of its parameter value back to the associated DMO
    * property.
    */
   private static class Scope
   implements OutputParameterAssigner
   {
      /** List of field assigners, created lazily */
      private List<FieldAssigner> list = null;
      
      /**
       * Discard any field assigners.
       */
      public void abort()
      {
         list = null;
      }
      
      /**
       * Cause each (if any) field assigner in the current scope to assign the
       * current value of its paramter value back to the DMO property with
       * which it is associated.
       */
      public void processAssignments()
      {
         if (list == null)
         {
            return;
         }
         
         boolean batch = (list.size() > 1);
         
         if (batch)
         {
            RecordBuffer.startBatch();
         }
         
         try
         {
            for (FieldAssigner next : list)
            {
               next.assign();
            }
         }
         finally
         {
            list = null;
            
            if (batch)
            {
               RecordBuffer.endBatch();
            }
         }
      }
      
      /**
       * Add a field assigner to this scope.
       * 
       * @param   assigner
       *          Field assigner.
       */
      private void addAssigner(FieldAssigner assigner)
      {
         if (list == null)
         {
            list = new ArrayList<FieldAssigner>(2);
            TransactionManager.registerOutputParameterAssigner(this);
         }
         
         list.add(assigner);
      }
   }
}
